from typing import Dict, List

import numpy as np
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from .compute_z import get_module_input_output_at_words
from .memit_hparams import MEMITAugHyperParams


def compute_ks(
    model: AutoModelForCausalLM,
    tok: AutoTokenizer,
    requests: Dict,
    hparams: MEMITAugHyperParams,
    layer: int,
    context_templates: List[str],
):
    all_edit_prompt = [
        item 
        for request, context_prompt in zip(requests, context_templates)
        for item in context_prompt + [request["prompt"]]
    ]


    all_edit_subject = [
        request["subject"]
        for request in requests
        for _ in range(len(context_templates[requests.index(request)]) + 1)
    ]

    layer_ks = get_module_input_output_at_words(
        model,
        tok,
        layer,
        context_templates=all_edit_prompt,
        words=all_edit_subject,
        module_template=hparams.rewrite_module_tmp,
        fact_token_strategy=hparams.fact_token,
    )[0]

    context_type_lens = [0] + [len(context_type) for context_type in context_templates]
    context_len = sum(context_type_lens)
    context_type_csum = np.cumsum(context_type_lens).tolist()

    ans = []
    for i in range(0, layer_ks.size(0), context_len):
        tmp = []
        for j in range(len(context_type_csum) - 1):
            start, end = context_type_csum[j], context_type_csum[j + 1]
            tmp.append(layer_ks[i + start : i + end].mean(0))
        ans.append(torch.stack(tmp, 0).mean(0))
    return torch.stack(ans, dim=0)
